/*
Copyright (C) 2004-2013 Jacquelin POTIER <jacquelin.potier@free.fr>
Dynamic aspect ratio code Copyright (C) 2004-2013 Jacquelin POTIER <jacquelin.potier@free.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

//-----------------------------------------------------------------------------
// Object: make an optimized output buffer (strcat like) function with reallocation if needed
//-----------------------------------------------------------------------------
#include "OutputString.h"
#include <stdio.h>

#pragma intrinsic (memcpy,memset,memcmp)

void COutputString::Prepend(TCHAR* StringToAdd,SIZE_T StringToAddSize)
{
    SIZE_T MinRequiredLen = this->Size + StringToAddSize+1;
    SIZE_T OldSize = this->Size;
    if ( MinRequiredLen > this->MaxSize)
    {
        // warning : != that Increase size

        // store current string buffer to TmpString
        TCHAR* TmpString = this->String;
        // double string size at least and assume that size is enough to store StringToAdd
        this->MaxSize = __max(this->MaxSize*2,MinRequiredLen*2);// *2 to reduce as possible the number of allocation / delete
        // allocate a new buffer
        this->String = new TCHAR[this->MaxSize];
        if (this->Size)
        {
            // copy new content to allocated string
            memcpy(this->String,StringToAdd,StringToAddSize*sizeof(TCHAR));
            // copy previous content to new allocated string
            memcpy(&this->String[StringToAddSize],StringToAdd,(this->Size+1)*sizeof(TCHAR));
        }

        // delete previously allocated buffer
        delete[] TmpString;

        this->Size+=StringToAddSize;

        return;
    }
    if (OldSize>0)
        // change old string position first
        memmove(&this->String[StringToAddSize],this->String,(this->Size+1)*sizeof(TCHAR));

    // copy new content
    memcpy(this->String,StringToAdd,StringToAddSize*sizeof(TCHAR));

    this->Size+=StringToAddSize;
}

void COutputString::Append(TCHAR* StringToAdd,SIZE_T StringToAddSize)
{
    if ( ( this->Size + StringToAddSize + 1) > this->MaxSize)
    {
        this->IncreaseCapacity(this->Size + StringToAddSize+1);// +1 for \0
    }
    // do the _tcscat
    // optimize as we know current string size
    // _tcscat(this->String,StringToAdd);
    // _tcscpy(&this->String[this->Size],StringToAdd);
    // as we know size optimize even more using memcpy
    memcpy(&this->String[this->Size],StringToAdd,StringToAddSize*sizeof(TCHAR));
    this->Size+=StringToAddSize;
    // assume string end
    this->String[this->Size]=0;
}

void COutputString::Append(TCHAR CharToAdd)
{  
    if (CharToAdd=='\0')
        return;

    if ( ( this->Size + 2) > this->MaxSize)// +2: +1 for \0 +1 for current char
    {
        this->IncreaseCapacity(this->Size + 2);
    }
    // do the _tcscat
    // // optimize as we know current string size
    // // _tcscat(this->String,StringToAdd);
    // _tcscpy(&this->String[this->Size],StringToAdd);
    // optimize : avoid _tcscpy call
    TCHAR* pAdd;
    pAdd = &this->String[this->Size];
    *pAdd = CharToAdd;
    pAdd++;
    *pAdd = 0;
    this->Size++;
}

void COutputString::AppendW(WCHAR CharToAdd)
{
#if (defined(UNICODE)||defined(_UNICODE))
    this->Append(CharToAdd);
#else
    CHAR ch;
    ::WideCharToMultiByte(CP_ACP, 0, &CharToAdd, 1, &ch, 1, NULL, NULL);
    this->Append(ch);
#endif
}
void COutputString::AppendW(WCHAR* UnicodeStringToAdd,SIZE_T StringToAddSize)
{
#if (defined(UNICODE)||defined(_UNICODE))
    this->Append(UnicodeStringToAdd,StringToAddSize);
#else
    // increase internal buffer and do direct output into internal buffer
    if ( ( this->Size + StringToAddSize + 1) > this->MaxSize)
    {
        this->IncreaseCapacity(this->Size + StringToAddSize+1);// +1 for \0
    }
    ::WideCharToMultiByte(CP_ACP, 0, UnicodeStringToAdd, StringToAddSize, &this->String[this->Size], StringToAddSize, NULL, NULL);
    this->Size+=StringToAddSize;
    // assume string end
    this->String[this->Size]=0;
#endif
}
void COutputString::AppendA(CHAR CharToAdd)
{
#if (defined(UNICODE)||defined(_UNICODE))
    WCHAR wch;
    ::MultiByteToWideChar(CP_ACP, 0, &CharToAdd, 1, &wch, 1);
    this->Append(wch);
#else
    this->Append(CharToAdd);
#endif
}
void COutputString::AppendA(CHAR* AsciiStringToAdd,SIZE_T StringToAddSize)
{
#if (defined(UNICODE)||defined(_UNICODE))
    // increase internal buffer and do direct output into internal buffer
    if ( ( this->Size + StringToAddSize + 1) > this->MaxSize)
    {
        this->IncreaseCapacity(this->Size + StringToAddSize+1);// +1 for \0
    }
    ::MultiByteToWideChar(CP_ACP, 0, AsciiStringToAdd, (int)StringToAddSize, &this->String[this->Size], (int)StringToAddSize);
    this->Size+=StringToAddSize;
    // assume string end
    this->String[this->Size]=0;
#else
    this->Append(AsciiStringToAdd,StringToAddSize);
#endif
}


void COutputString::IncreaseCapacity(SIZE_T MinRequiredLen)
{
    if (MinRequiredLen<this->MaxSize)
        return;

    // store current string buffer to TmpString
    TCHAR* TmpString = this->String;
    // double string size at least and assume that size is enough to store StringToAdd
    this->MaxSize = __max(this->MaxSize*2,MinRequiredLen*2);// *2 to reduce as possible the number of allocation / delete
    // allocate a new buffer
    this->String = new TCHAR[this->MaxSize];
    // copy previous content to new allocated string
    // optimize : avoid checking of \0
    // _tcscpy(this->String,TmpString);
    if (this->Size)
        memcpy(this->String,TmpString,(this->Size+1)*sizeof(TCHAR));
    else
        *this->String=0;

    // delete previously allocated buffer
    delete[] TmpString;
}

int COutputString::Snprintf(SIZE_T MaxCount,const TCHAR* Format,...)
{
    va_list args;
    va_start(args, Format);
    SIZE_T MinSize = MaxCount+1;
    if (this->MaxSize<MinSize)
        this->IncreaseCapacity(MinSize);
    int RetValue = _vsntprintf(this->String,MaxCount,Format,args);
    // _vsnprintf and _vsnwprintf return the number of characters written, not including the terminating null character, or a negative value if an output error occurs
    // If the number of characters to write exceeds count, then count characters are written and 1 is returned
    if (RetValue>0)
        this->Size = (SIZE_T)RetValue;
    else
    {
        this->String[MaxCount-1]=0;
        this->Size = _tcslen(this->String);
    }
    va_end(args);
    return RetValue;
}

int COutputString::AppendSnprintf(SIZE_T MaxCount,const TCHAR* Format,...)
{
    va_list args;
    va_start(args, Format);
    int RetValue = this->AppendVsnprintf(MaxCount,Format,args);
    va_end(args);
    return RetValue;
}

int COutputString::AppendVsnprintf(SIZE_T MaxCount,const TCHAR* Format,va_list args)
{
    SIZE_T MinSize = this->Size+MaxCount+1;
    if (this->MaxSize<MinSize)
        this->IncreaseCapacity(MinSize);
    // take pointer after IncreaseCapacity call as IncreaseCapacity can affect this->String
    TCHAR* Ptr = this->String+this->Size;
    int RetValue = _vsntprintf(Ptr,MaxCount,Format,args);
    // _vsnprintf and _vsnwprintf return the number of characters written, not including the terminating null character, or a negative value if an output error occurs
    // If the number of characters to write exceeds count, then count characters are written and 1 is returned
    if (RetValue>0)
        this->Size+= (SIZE_T)RetValue;
    else
    {
        Ptr[MaxCount-1]=0;
        this->Size = _tcslen(this->String);
    }
    return RetValue;
}

//-----------------------------------------------------------------------------
// Name: Ellipse
// Object: Ellipse current string representation
// Parameters :
//      SIZE_T MaxWantedNbChars : max number of char required including \0
// Return : return TRUE is Ellipse has been done, FALSE is current size is less than MaxWantedNbChars
//-----------------------------------------------------------------------------
BOOL COutputString::Ellipse(SIZE_T MaxWantedNbChars)
{
    // if string size is smaller than required char numbers
    if ( this->Size < MaxWantedNbChars)
        return FALSE;
    // else: (this->Size > MaxWantedNbChars)

    // if possible, add "..." at the end of string to tell user string representation is not fully completed
    if (MaxWantedNbChars>3)
    {
        TCHAR* pc = &this->String[MaxWantedNbChars-4];
        _tcscpy(pc,_T("..."));
    }
    else
    {
        this->String[MaxWantedNbChars-1]=0;
    }
    this->Size = MaxWantedNbChars-1;// -1 \0 is not taken into account for length

    return TRUE;
}